"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from .subclass import get_subclasses
from .flatiter import flatten, concatenate
from .iterhelp import get_any_value, merge_dict
from .importhelp import can_import
from .network import download_file
from .timer import Timer
from .critical import CriticalSection
from . import idx
from . import mergetools
from .environtools import EnvironmentalVariable
from .contexttools import redirect_stderr

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF
